# -*- coding: utf-8 -*-

import sys
import json
import copy

from rich.table import Table
from rich.console import Console
from suite_py.lib import logger
from suite_py.lib.handler import prompt_utils
from suite_py.lib.handler.qainit_handler import QainitHandler
from suite_py.lib.handler import git_handler as git
from suite_py.lib.handler.git_handler import GitHandler


class QA:
    def __init__(self, action, project, config, tokens, flags=None):
        self._action = action
        self._project = project
        self._flags = flags
        self._config = config
        self._tokens = tokens
        self._git = GitHandler(project, config)
        self._qainit = QainitHandler(project, config, tokens)

    def run(self):
        if self._action == "list":
            self._list()
        elif self._action == "create":
            self._create()
        elif self._action == "update":
            self._update()
        elif self._action == "delete":
            self._delete()
        elif self._action == "check":
            self._check()
        elif self._action == "describe":
            self._describe()
        elif self._action == "update-quota":
            self._update_quota()

    def _check(self):
        r = self._qainit.execute("GET", "/api/v1/user")
        logger.info(
            "Checking configuration. If there is an issue, check ~/.suite_py/config.yml file and execute: suite-py login"
        )
        logger.debug(json.dumps(r.json(), indent=2))

    def _list(self):
        # init empty table with column (useful for reset)
        empty_table = Table()
        empty_table.add_column("Name", style="purple")
        empty_table.add_column("Hash", style="green")
        empty_table.add_column("Created by", style="white")
        empty_table.add_column("Updated by", style="white")
        empty_table.add_column("Deleted by", style="white")
        empty_table.add_column("Status", style="white")
        table = copy.deepcopy(empty_table)
        console = Console()

        # add filtering
        filters = []
        status_values = '"created","creating","updated","updating","failed"'
        if len(self._flags["status"]) > 0:
            status_values = ",".join(
                [f'"{status}"' for status in self._flags["status"]]
            )
        filters.append(f"status=[{status_values}]")

        if self._flags["user"] is not None:
            filters.append(f"user={self._flags['user']}")
        filters_string = "&".join(filters)

        # execute query with pagination and filtering
        page_number = 1
        while True:
            r = self._qainit.execute(
                "GET",
                f'/api/v1/qa?{filters_string}&page_size={self._config.qainit["table_size"]}&page={page_number}',
            )
            response = r.json()
            qa_list = response["list"]
            for qa in qa_list:
                table.add_row(
                    qa["name"],
                    qa["hash"],
                    qa.get("created", {}).get("github_username", "/")
                    if qa["created"] is not None
                    else "/",
                    qa.get("updated", {}).get("github_username", "/")
                    if qa["updated"] is not None
                    else "/",
                    qa.get("deleted", {}).get("github_username", "/")
                    if qa["deleted"] is not None
                    else "/",
                    qa["status"],
                )
            console.print(table)

            # break conditions
            if response["page_number"] >= response["total_pages"]:
                break
            if not prompt_utils.ask_confirm(
                f"I found {response['total_entries']} results. Do you want to load a few more?",
                False,
            ):
                break
            page_number += 1
            # table reset
            table = copy.deepcopy(empty_table)

    def _describe(self):
        qa_hash = self._flags["qa_hash"]
        jsonify = self._flags["json"]

        table = Table()
        table.add_column("Microservice", style="purple", no_wrap=True)
        table.add_column("Branch", style="green")
        table.add_column("Drone build")
        table.add_column("Status", style="white")
        console = Console()

        r = self._qainit.execute(
            "GET",
            f"/api/v1/qa/{qa_hash}",
        )
        try:
            resources_list = r.json()["list"]["resources"]
            for resource in resources_list:
                if resource["type"] == "microservice":
                    drone_url = (
                        "[blue][u]"
                        + "https://drone-1.prima.it/primait/"
                        + resource["name"]
                        + "/"
                        + resource["promoted_build"]
                        + "[/u][/blue]"
                    )
                    table.add_row(
                        resource["name"],
                        resource["ref"],
                        drone_url,
                        resource["status"],
                    )
            if jsonify:
                print(json.dumps(r.json(), indent=2))
            else:
                print(
                    f"""
> Prima url: https://www-{qa_hash}.prima.qa
> Backoffice (Borat) url: https://backoffice-{qa_hash}.prima.qa
> Urania url: http://urania-{qa_hash}.prima.qa
> Bburago url: http://bburago-{qa_hash}.prima.qa
> Ermes url: http://ermes-{qa_hash}.prima.qa
> Hal9000 url: http://hal9000{qa_hash}.prima.qa
> Fidaty url: http://fidaty-{qa_hash}.prima.qa
> Peano url: http://peano-{qa_hash}.prima.qa
> Assange url: https://assange-{qa_hash}.prima.qa
> Activia url: http://activia-{qa_hash}.prima.qa
> Skynet url: http://skynet-{qa_hash}.prima.qa
> Roger url: http://roger-{qa_hash}.prima.qa
> Leftorium url: http://leftoriu-{qa_hash}.prima.qa
> Rachele url: http://rachele-{qa_hash}.prima.qa
> Maia App url: https://api-{qa_hash}.prima.qa
> Maia Intermediari url: https://api-intermediari-{qa_hash}.prima.qa
> Maia Casa url: https://api-casa-{qa_hash}.prima.qa
> Maia Pagamenti url: https://api-pagamenti-{qa_hash}.prima.qa
> Maia Sostituzioni url: https://api-sostituzioni-{qa_hash}.prima.qa
> Legion url: http://legion-{qa_hash}.prima.qa
> Crash url: https://crash-{qa_hash}.prima.qa
> Starsky url: https://starsky-{qa_hash}.prima.qa
> Hutch url: https://hutch-{qa_hash}.prima.qa
> Vianello url: https://vianello-{qa_hash}.prima.qa
> Domus url: https://domus-{qa_hash}.prima.qa
> Lira url: https://lira-{qa_hash}.prima.qa
> Baggio url: https://baggio-{qa_hash}.prima.qa
> Mario url: https://mario-{qa_hash}.prima.qa
> Mario Backend url: https://api-mario-{qa_hash}.prima.qa
> Zuhause url: https://zuhause-{qa_hash}.prima.qa
> Pyxis url: https://pyxis-{qa_hash}.prima.qa
> Caritas url: https://caritas-{qa_hash}.prima.qa
> Cashabck url: https://cashback-{qa_hash}.prima.qa
> Evvivass url: https://evvivass-{qa_hash}.prima.qa
                """
                )
                console.print(table)
        except TypeError:
            logger.error("Wrong hash")

    def _delete(self):
        qa_hash = self._flags

        r = self._qainit.execute(
            "DELETE",
            f"/api/v1/qa/{qa_hash}",
        )
        logger.info("QA deletion initiated")
        logger.debug(json.dumps(r.json(), indent=2))

    def _create(self):
        r = self._qainit.execute("GET", "/api/v1/user")
        r = r.json()
        if not r["quota"]["remaining"] > 0:
            logger.error("There's no remaining quota for you.")
            sys.exit("-1")
        if "staging" in self._qainit.url:
            qa_default_name = (
                f"staging_{git.get_username()}_{self._git.current_branch_name()}"
            )
        else:
            qa_default_name = f"{git.get_username()}_{self._git.current_branch_name()}"

        qa_name = prompt_utils.ask_questions_input(
            "Choose the QA name: ", default_text=qa_default_name
        )

        srv_list = self._qainit.create_services_body(prj_list=self._flags)

        body = {"name": f"{qa_name}", "services": srv_list}
        logger.debug(json.dumps(body))
        r = self._qainit.execute(
            "POST",
            "/api/v1/qa",
            body=json.dumps(body),
        )
        logger.info("QA creation initiated. Your namespace hash: ")
        logger.info(r.json()["hash"])
        logger.debug(json.dumps(r.json(), indent=2))

    def _update(self):
        qa_hash = self._flags[0]
        prj_list = self._flags[1:]

        srv_list = self._qainit.create_services_body(prj_list)

        body = {"services": srv_list}
        logger.debug(json.dumps(body))
        r = self._qainit.execute(
            "PUT",
            f"/api/v1/qa/{qa_hash}",
            body=json.dumps(body),
        )
        logger.info("QA update initiated")
        logger.debug(json.dumps(r.json(), indent=2))

    def _update_quota(self):
        username = prompt_utils.ask_questions_input("Insert GitHub username: ")
        quota = prompt_utils.ask_questions_input("Insert new quota value: ")

        body = {"github_username": f"{username}", "quota": f"{quota}"}
        logger.debug(json.dumps(body))
        r = self._qainit.execute(
            "POST",
            "/api/v1/user/quota",
            body=json.dumps(body),
        )

        logger.info("Quota updated.")
        logger.debug(json.dumps(r.json(), indent=2))



# -*- coding: utf-8 -*-
import sys
import re

from halo import Halo

from suite_py.lib import logger
from suite_py.lib.handler.captainhook_handler import CaptainHook
from suite_py.lib.handler import prompt_utils


class Aggregator:
    def __init__(self, config, show_list, change):
        self._captainhook = CaptainHook(config)
        self._show_list = show_list
        self._change = change

    def run(self):
        if self._show_list:
            self._list_aggregators()
            return

        if self._change:
            aggregator = self._select_aggregator()
            address = prompt_utils.ask_questions_input(
                "Insert QA url or press enter to set staging URL: ",
                default_text="staging.prima.it",
            )

            if address.startswith("http"):
                address = re.sub(r"https?://?", "", address)

            change_request = self._captainhook.change_aggregator(aggregator, address)
            self._handle_captainhook_response(change_request, aggregator, address)

    def _handle_captainhook_response(self, request, aggregator, address):
        if request.status_code == 200:
            change_request = request.json()
            if change_request["success"]:
                logger.info(f"CNAME updated! Now {aggregator} is pointing to {address}")
            else:
                cases = {
                    "cloudflare_error": "Error during Cloudflare invocation.",
                    "unknown_dns_record": "Impossible to find DNS record associated with aggregator.",
                    "unknown_aggregator": "Aggregator not found.",
                    "invalid_qa_address": "The QA address does not meet the requirements.",
                }
                logger.error(cases.get(change_request["error"], "unknown error"))
        else:
            logger.error("An error has occurred on Captainhook.")

    def _select_aggregator(self):
        with Halo(text="Loading aggregators...", spinner="dots", color="magenta"):
            choices = [
                {"name": agg["name"], "value": agg["name"]}
                for agg in self._captainhook.aggregators().json()
            ]
        if choices:
            choices.sort(key=lambda x: x["name"])
            return prompt_utils.ask_choices("Select aggregator: ", choices)

        logger.error("There are no aggregators on Captainhook.")
        sys.exit(-1)

    def _list_aggregators(self):
        with Halo(text="Loading...", spinner="dots", color="magenta"):
            aggregators = self._captainhook.aggregators()

        if aggregators.status_code != 200:
            logger.error("Unable to retrieve the list of aggregators.")
            return

        message = "\n".join(
            [f"* {a['name']:>20} => {a['content']:>32}" for a in aggregators.json()]
        )

        logger.info(f"\nAvailable aggregators:\n{message}\n")



# -*- encoding: utf-8 -*-
import sys

from github import GithubException

from suite_py.commands.ask_review import AskReview
from suite_py.lib import logger
from suite_py.lib.handler import prompt_utils
from suite_py.lib.handler.git_handler import GitHandler, get_commit_logs
from suite_py.lib.handler.github_handler import GithubHandler
from suite_py.lib.handler.youtrack_handler import YoutrackHandler


class OpenPR:
    def __init__(self, project, config, tokens):
        self._project = project
        self._config = config
        self._tokens = tokens
        self._youtrack = YoutrackHandler(config, tokens)
        self._git = GitHandler(project, config)
        self._branch_name = self._git.current_branch_name()
        self._github = GithubHandler(tokens)

    def run(self):
        if not self._git.remote_branch_exists(self._branch_name):
            logger.warning(f"No branch named {self._branch_name} found on GitHub")
            if prompt_utils.ask_confirm(
                "Do you want to commit all the files and push them?"
            ):
                self._git.add()
                self._git.commit("Initial commit")
                self._git.push(self._branch_name)
            else:
                logger.error("Please, run 'git push' manually")
                sys.exit(-1)

        pulls = self._github.get_pr_from_branch(self._project, self._branch_name)
        if pulls.totalCount:
            pr = pulls[0]
            logger.info(
                f"There is a pull request on GitHub for the branch {self._branch_name}"
            )

            if prompt_utils.ask_confirm(
                "Do you want to change the description of the pull request?"
            ):
                self._edit_pr(pr)
            sys.exit(0)

        youtrack_id = self._get_youtrack_id()

        self._create_pr(youtrack_id)

    def _get_youtrack_id(self):
        youtrack_id = self._youtrack.get_card_from_name(self._branch_name)
        if youtrack_id:
            return youtrack_id

        logger.warning(
            "Couldn't find a YouTrack issue in the branch name or the selected issue does not exist"
        )
        if prompt_utils.ask_confirm(
            "Do you want to link the pull request with an issue?"
        ):
            return self._ask_for_card_id()
        return None

    def _ask_for_card_id(self):
        card_id = prompt_utils.ask_questions_input("Enter card ID (ex: PRIMA-1234): ")
        if self._youtrack.validate_issue(card_id):
            return card_id
        logger.error("ID does not exist on YouTrack")
        return self._ask_for_card_id()

    def _create_pr(self, youtrack_id):
        if youtrack_id:
            logger.info(
                f"Creating pull request on the {self._project} project for the {self._branch_name} branch linked with the {youtrack_id} card"
            )
            link = self._youtrack.get_link(youtrack_id)
            title = (
                f"[{youtrack_id}]: {self._youtrack.get_issue(youtrack_id)['summary']}"
            )
        else:
            logger.warning(
                f"Creating pull request on the {self._project} project for the {self._branch_name} branch NOT linked to YouTrack card"
            )
            link = ""
            title = _ask_for_title()

        base_branch = self._ask_for_base_branch()
        body = self._ask_for_description("", base_branch=base_branch, link=link)

        is_draft = prompt_utils.ask_confirm(
            "Do you want to open the pull request as a draft?", default=False
        )

        try:
            pr = self._github.create_pr(
                self._project, self._branch_name, title, body, base_branch, is_draft
            )
            logger.info(f"Pull request with number {pr.number} created! {pr.html_url}")
        except GithubException as e:
            logger.error("Error during GitHub invocation: ")
            logger.error(e.data["errors"][0])
            sys.exit(-1)

        if youtrack_id:
            self._youtrack.comment(youtrack_id, f"PR {self._project} -> {pr.html_url}")
            logger.info(f"Added the pull request link in the card {youtrack_id}")

        if prompt_utils.ask_confirm("Do you want to insert reviewers?"):
            AskReview(self._project, self._config, self._tokens).run()

    def _edit_pr(self, pr):
        pr_body = self._ask_for_description(pr.body)
        pr.edit(body=pr_body)
        logger.info("Pull request modified")

    def _ask_for_base_branch(self):
        branch = prompt_utils.ask_questions_input(
            "Enter the base branch of the pull request: ", "master"
        )
        return branch

    def _ask_for_description(self, pr_body, **opts):
        link = opts.get("link")
        base_branch = opts.get("base_branch")
        if pr_body == "":
            if link:
                pr_body = link + "\n\n"
            try:
                with open("pull_request_template.md", "r") as f:
                    pr_body = pr_body + f.read() + "\n"
            except Exception:
                pass
            if base_branch and self._config.user["use_commits_in_pr_body"]:
                commit_list = get_commit_logs(base_branch)
                commit_list.reverse()
                pr_body = pr_body + "\n---\n".join(commit_list)
                input(
                    "Enter the description of the pull request. Pressing enter will open the default editor"
                )
        description = prompt_utils.ask_questions_editor(
            question_text="Enter the description of the PR: ", default_text=pr_body
        )
        if description == "":
            logger.warning("The description of the pull request cannot be empty")
            return self._ask_for_description(pr_body)
        return description


def _ask_for_title():
    title = prompt_utils.ask_questions_input("Enter the title of the pull request: ")
    if title == "":
        logger.warning("The title of the pull request cannot be empty")
        return _ask_for_title()
    return title



# -*- coding: utf-8 -*-
import subprocess
import sys
import yaml

from halo import Halo

from suite_py.lib import logger
from suite_py.lib.handler import git_handler as git
from suite_py.lib.handler import prompt_utils
from suite_py.lib.handler.drone_handler import DroneHandler
from suite_py.lib.handler.git_handler import GitHandler
from suite_py.lib.handler.github_handler import GithubHandler
from suite_py.lib.handler.youtrack_handler import YoutrackHandler
from suite_py.lib.qainit import get_qa_projects, qainit_deploy


class CreateQA:
    def __init__(self, project, config, tokens):
        self._project = project
        self._config = config
        self._youtrack = YoutrackHandler(config, tokens)
        self._git = GitHandler(project, config)
        self._github = GithubHandler(tokens)
        self._qainit_drone = DroneHandler(config, tokens, repo="qainit")

    def run(self):
        qa_projects = get_qa_projects(self._config)
        if self._project not in qa_projects:
            logger.error(
                f"The {self._project} project is not part of the QA environments, nothing to do"
            )
            sys.exit(0)

        with Halo(text="Updating prima-twig...", spinner="dots", color="magenta"):
            try:
                subprocess.run(
                    ["gem", "update", "prima-twig"],
                    check=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
            except Exception:
                logger.error(
                    "Error updating prima-twig, please run `gem update prima-twig`"
                )
                sys.exit(-1)

        branch_name = self._git.current_branch_name()
        youtrack_id = self._youtrack.get_card_from_name(branch_name)

        if youtrack_id:
            if prompt_utils.ask_confirm(
                f"Are there any other projects to add to the QA that refer to the card {youtrack_id}?",
                default=False,
            ):
                twig_arg = self._multi_repo_qa(youtrack_id, qa_projects)
            else:
                twig_arg = self._mono_repo_qa(branch_name)
        else:
            logger.warning(
                "I couldn't find a YouTrack issue linked to the branch, it is not possible to automatically link other projects to this QA"
            )
            logger.warning(
                "If you want to create a QA with different branches on different projects please use twig"
            )
            twig_arg = self._mono_repo_qa(branch_name)

        qainit_deploy(twig_arg, self._config)

        if youtrack_id:
            logger.info("Updating card on YouTrack...")
            self._youtrack.update_state(
                youtrack_id, self._config.youtrack["test_state"]
            )

        drone_url = self._qainit_drone.get_last_build_url(git.get_username())
        if drone_url:
            logger.info(f"You can follow the creation of the QA at {drone_url}")

        logger.info("QA setup successful!")

    def _multi_repo_qa(self, youtrack_id, qa_projects):
        logger.info("Looking for the other branches on GitHub...")
        selected_repos = []
        for repo in [self._github.get_repo(p) for p in qa_projects]:
            print(".", end="", flush=True)
            for branch in repo.get_branches():
                if youtrack_id in branch.name:
                    selected_repos.append((repo.name, branch))

        print(".")
        logger.info("Projects found:")
        for elem in selected_repos:
            logger.info(f"{elem[0]} --> {elem[1].name}")

        return _twigify(selected_repos)

    def _mono_repo_qa(self, branch_name):
        logger.info(
            f"Creating QA with the project {self._project} and branch {branch_name}"
        )
        gh_branch = self._github.get_repo(self._project).get_branch(branch_name)
        selected_repos = [(self._project, gh_branch)]

        return _twigify(selected_repos)


def _twigify(projects_list):
    projects_dict = {}
    for project in projects_list:
        committer = ""

        try:
            committer = project[1].commit.committer.email or ""
        except Exception:
            committer = project[1].commit.commit.committer.email or ""

        projects_dict[project[0]] = {
            "name": project[1].name,
            "revision": project[1].commit.sha[:15],
            "committer": committer,
            "default_branch": False,
        }

    return yaml.dump(projects_dict, explicit_start=True, default_flow_style=False)



# -*- encoding: utf-8 -*-
from itertools import groupby
import os
import re
import subprocess
from subprocess import CalledProcessError
import sys
import semver

from halo import Halo

from suite_py.lib import logger


class GitHandler:
    def __init__(self, repo, config):
        self._repo = repo
        self._projects_home = config.user["projects_home"]
        self._path = os.path.join(self._projects_home, repo)

    def get_repo(self):
        return self._repo

    def get_path(self):
        return self._path

    def clone(self):
        subprocess.run(
            ["git", "clone", f"git@github.com:primait/{self._repo}.git"],
            cwd=self._projects_home,
            check=True,
        )

    def checkout(self, branch, new=False):
        try:
            if new:
                subprocess.run(
                    ["git", "checkout", "-b", branch],
                    cwd=self._path,
                    check=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
            else:
                subprocess.run(
                    ["git", "checkout", branch],
                    cwd=self._path,
                    check=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
                if not self.is_detached():  # Skip pulling if detached
                    self.pull(rebase=True)
        except (CalledProcessError) as e:
            logger.error(f"Error during command execution: {e}")
            sys.exit(-1)

    def commit(self, commit_message="", dummy=False):
        try:
            if dummy:
                subprocess.run(
                    ["git", "commit", "--allow-empty", "-m", commit_message],
                    cwd=self._path,
                    check=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
            else:
                subprocess.run(
                    ["git", "commit", "-am", commit_message],
                    cwd=self._path,
                    check=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
        except (CalledProcessError) as e:
            logger.error(f"Error during command execution: {e}")
            sys.exit(-1)

    def add(self):
        try:
            subprocess.run(["git", "add", "."], cwd=self._path, check=True)
        except (CalledProcessError) as e:
            logger.error(f"Error during command execution: {e}")
            sys.exit(-1)

    def push(self, branch, remote="origin"):
        with Halo(text=f"Pushing {self._repo}...", spinner="dots", color="magenta"):
            try:
                subprocess.run(
                    ["git", "push", remote, branch],
                    cwd=self._path,
                    check=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
            except (CalledProcessError) as e:
                logger.error(f"Error during command execution: {e}")
                sys.exit(-1)

    def pull(self, rebase=False):
        with Halo(text=f"Pulling {self._repo}...", spinner="dots", color="magenta"):
            try:
                if rebase:
                    subprocess.run(
                        ["git", "pull", "--rebase"],
                        cwd=self._path,
                        check=True,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                    )
                else:
                    subprocess.run(
                        ["git", "pull"],
                        cwd=self._path,
                        check=True,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                    )
            except (CalledProcessError) as e:
                logger.error(f"Error during command execution: {e}")
                sys.exit(-1)

    def fetch(self, remote="origin"):
        with Halo(text=f"Fetching {self._repo}...", spinner="dots", color="magenta"):
            try:
                subprocess.run(
                    ["git", "fetch", "--quiet"],
                    cwd=self._path,
                    check=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
                subprocess.run(
                    ["git", "fetch", "-p", remote, "--quiet"],
                    cwd=self._path,
                    check=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
            except (CalledProcessError) as e:
                logger.error(f"Error during command execution: {e}")
                sys.exit(-1)

    def sync(self):
        self.fetch()
        self.checkout("master")
        self.pull()

    def check_repo_cloned(self):
        if self._repo not in os.listdir(self._projects_home):
            logger.warning("The project is not in your home path, cloning it now...")
            self.clone()

    def delete_remote_branch(self, branch):
        try:
            self.push(f":{branch}")
            return True
        except BaseException:
            return False

    def get_last_tag(self):
        self.sync()
        return (
            subprocess.check_output(["git", "describe", "--abbrev=0", "--tags"])
            .decode("utf-8")
            .strip()
        )

    def search_remote_branch(self, regex):
        try:
            output = subprocess.run(
                ["git", "branch", "-r", "--list", regex],
                cwd=self._path,
                check=True,
                stdout=subprocess.PIPE,
            )

            result = output.stdout.decode("utf-8").strip("\n").strip()

            if result.startswith("origin/"):
                return result[7:]
            return result

        except CalledProcessError:
            return ""

    def local_branch_exists(self, branch):
        try:
            subprocess.run(
                ["git", "show-ref", "--quiet", f"refs/heads/{branch}"],
                cwd=self._path,
                check=True,
            )
            return True
        except CalledProcessError:
            return False

    def remote_branch_exists(self, branch):
        try:
            self.fetch()
            subprocess.run(
                ["git", "show-ref", "--quiet", f"refs/remotes/origin/{branch}"],
                cwd=self._path,
                check=True,
            )
            return True
        except CalledProcessError:
            return False

    def current_branch_name(self):
        try:
            output = subprocess.run(
                ["git", "rev-parse", "--abbrev-ref", "HEAD"],
                cwd=self._path,
                stdout=subprocess.PIPE,
                check=True,
            ).stdout.decode("utf-8")
            return re.sub("\n", "", output)
        except (CalledProcessError) as e:
            logger.error(f"Error during command execution: {e}")
            sys.exit(-1)

    def is_detached(self):
        return self.current_branch_name() == "HEAD"

    def is_dirty(self, untracked=False):
        if untracked:
            command = ["git", "status", "--porcelain"]
        else:
            command = ["git", "status", "--porcelain", "-uno"]

        try:
            output = subprocess.run(
                command,
                cwd=self._path,
                check=True,
                stdout=subprocess.PIPE,
            ).stdout.decode("utf-8")

            return len(output) != 0
        except (CalledProcessError) as e:
            logger.error(f"Error during command execution: {e}")
            sys.exit(-1)

    def reset(self):
        try:
            subprocess.run(
                ["git", "reset", "HEAD", "--hard", "--quiet"],
                cwd=self._path,
                check=True,
            )
        except CalledProcessError as e:
            logger.error(f"Error during git reset on {self._repo}: {e}")
            sys.exit(-1)


def is_repo(directory):
    try:
        subprocess.run(
            ["git", "status"],
            cwd=directory,
            check=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        return True
    except CalledProcessError:
        return False


def get_username():
    try:
        output = subprocess.run(
            ["git", "config", "user.name"], check=True, stdout=subprocess.PIPE
        ).stdout.decode("utf-8")
        output = re.sub("[^A-Za-z]", "", output)
        return re.sub("\n", "", output)
    except CalledProcessError:
        return None


def files_changed_between_commits(from_commit, to_commit):
    p = subprocess.Popen(  # pylint: disable=consider-using-with
        ["git", "diff", from_commit, to_commit, "--name-only"], stdout=subprocess.PIPE
    )
    result = p.communicate()[0]
    return result.decode("utf-8").splitlines()


def migrations_found(files_changed):
    for file in files_changed:
        if "migration" in file.lower():
            return True
    return False


def get_last_tag_number(tags):
    for tag in tags:
        if is_semver(tag.name):
            return tag.name
    return None


def is_semver(tag):
    try:
        return semver.parse(tag)
    except Exception:
        return None


def get_commit_logs(base_branch):
    """
    Get all the commits from HEAD that are *not* in `base_branch` and return
    their logs as a list. The commits are orderded from latest to first.
    """
    commits = subprocess.check_output(
        ["git", "log", "--pretty=medium", f"HEAD...{base_branch}"]
    ).decode("utf-8")

    commits_list = [
        "\n".join([line.lstrip() for line in split_result])
        for is_commit_line, split_result in groupby(
            commits.split("\n"),
            lambda line: re.match("commit [0-9a-f]{40}", line) is not None,
        )
        if not is_commit_line
    ]

    return commits_list



# -*- encoding: utf-8 -*-
import re

from youtrack.connection import Connection
from youtrack import YouTrackException


REGEX = r"([A-Za-z]+-[0-9]+)"
URL = "https://prima-assicurazioni-spa.myjetbrains.com/youtrack/issue/"


class YoutrackHandler:
    def __init__(self, config, tokens):
        self._client = Connection(url=config.youtrack["url"], token=tokens.youtrack)

    def get_projects(self):
        return self._client.getProjects()

    def get_users(self):
        return self._client.getUsers()

    def get_issue(self, issue_id):
        return self._client.getIssue(issue_id)

    def get_comments(self, issue_id):
        return self._client.getComments(issue_id)

    def update_deployed_field(self, issue_id):
        return self._client.executeCommand(issue_id, "Deployed Today")

    def validate_issue(self, issue_id):
        try:
            if self.get_issue(issue_id):
                return True
        except YouTrackException:
            pass
        return False

    def comment(self, issue_id, comment):
        self._client.executeCommand(issue_id, "comment", comment=comment)

    def update_state(self, issue_id, status):
        self._client.executeCommand(issue_id, f"State {status}")

    def add_tag(self, issue_id, label):
        self._client.executeCommand(issue_id, f"tag {label}")

    def assign_to(self, issue_id, user):
        self._client.executeCommand(issue_id, f"Assignee {user}")

    def get_link(self, issue_id):
        return f"{URL}{issue_id}"

    def get_issue_ids(self, commits):
        issue_ids = []
        for c in commits:
            issue_id = self.get_card_from_name(c.commit.message)
            if issue_id:
                issue_ids.append(issue_id)
        return issue_ids

    def get_card_from_name(self, name):
        if re.search(REGEX, name):
            id_card = re.findall(REGEX, name)[0]
            if self.validate_issue(id_card):
                return id_card
        return None

    def get_ids_from_release_body(self, body):
        return list(set(re.findall(REGEX, body)))


def replace_card_names_with_md_links(text):
    return re.sub(REGEX, f"[\\1]({URL}\\1)", text)
